@@ -1,4 +1,6 @@
-{-# LANGUAGE CPP, TupleSections, DeriveFunctor, GeneralizedNewtypeDeriving #-}
+{-# LANGUAGE CPP, LambdaCase, TupleSections, DeriveTraversable #-}
+{-# LANGUAGE GeneralizedNewtypeDeriving #-}
+{-# OPTIONS_GHC -fno-warn-unused-imports #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module     : Control.Selective
@@ -15,7 +17,7 @@
 -----------------------------------------------------------------------------
 module Control.Selective (
     -- * Type class
-    Selective (..), (<*?), branch, selectA, apS, selectM,
+    Selective (..), (<*?), branch, selectA, selectT, apS, selectM,
 
     -- * Conditional combinators
     ifS, whenS, fromMaybeS, orElse, andAlso, untilRight, whileS, (<||>), (<&&>),
@@ -139,6 +141,31 @@ import qualified Control.Monad.Trans.Writer.Strict as S
 class Applicative f => Selective f where
     select :: f (Either a b) -> f (a -> b) -> f b
 
+{- Why do we have skew associativity, where we can reassociate effects to the
+   left but not to the right?
+
+   The following two tables, which list all possible combinations of effect
+   execution and skipping, might give you some intuition on why this happens.
+
+     ---------------
+     (x <*? y) <*? z
+     ---------------
+      1     0      0
+      1     1      0
+      1     0      1
+      1     1      1
+
+     ---------------
+     x <*? (y <*? z)
+     ---------------
+     1      0     0
+     1      1     0
+     1      1     1
+
+   A key observation is that when effects are associated to the right, we can't
+   skip the effect y and execute the effect z: combination 101 is impossible.
+-}
+
 -- | An operator alias for 'select', which is sometimes convenient. It tries to
 -- follow the notational convention for 'Applicative' operators. The angle
 -- bracket pointing to the left means we always use the corresponding value.
@@ -170,6 +197,14 @@ branch x l r = fmap (fmap Left) x <*? fmap (fmap Right) l <*? r
 selectA :: Applicative f => f (Either a b) -> f (a -> b) -> f b
 selectA x y = (\e f -> either f id e) <$> x <*> y
 
+-- | If a functor is both 'Applicative' and 'Traversable', we can implement
+-- 'select' in another interesting way: the effects associated with the second
+-- argument can be skipped as long as the first argument contains only 'Right's.
+selectT :: (Applicative f, Traversable f) => f (Either a b) -> f (a -> b) -> f b
+selectT x y = case sequenceA x of
+    Left  a  -> ($a) <$> y
+    Right fb -> fb
+
 {-| Recover the application operator '<*>' from 'select'. /Rigid/ selective
 functors satisfy the law '<*>' @=@ 'apS' and furthermore, the resulting
 applicative functor satisfies all laws of 'Applicative':
@@ -196,8 +231,8 @@ apS f x = select (Left <$> f) ((&) <$> x)
 -- | One can easily implement a monadic 'selectM' that satisfies the laws,
 -- hence any 'Monad' is 'Selective'.
 selectM :: Monad f => f (Either a b) -> f (a -> b) -> f b
-selectM x y = x >>= \e -> case e of Left  a -> ($a) <$> y -- execute y
-                                    Right b -> pure b     -- skip y
+selectM x y = x >>= \case Left  a -> ($a) <$> y -- execute y
+                          Right b -> pure b     -- skip y
 
 -- Many useful 'Monad' combinators can be implemented with 'Selective'
 
@@ -240,7 +275,7 @@ matchS (Cases cs _) x f = foldr (\c -> eliminate c (f c)) (Left <$> x) cs
 matchM :: Monad m => Cases a -> m a -> (a -> m b) -> m (Either a b)
 matchM (Cases _ p) mx f = do
     x <- mx
-    if p x then Right <$> (f x) else return (Left x)
+    if p x then Right <$> f x else return (Left x)
 
 -- TODO: Add a type-safe version based on @KnownNat@.
 -- | A restricted version of monadic bind. Fails with an error if the 'Bounded'
@@ -367,17 +402,19 @@ instance Monoid m => Applicative (Over m) where
     pure _            = Over mempty
     Over x <*> Over y = Over (mappend x y)
 
+-- select = selectA
 instance Monoid m => Selective (Over m) where
     select (Over x) (Over y) = Over (mappend x y)
 
 -- | Static analysis of selective functors with under-approximation.
 newtype Under m a = Under { getUnder :: m }
-    deriving (Eq, Functor, Ord, Show)
+    deriving (Eq, Functor, Ord, Show, Foldable, Traversable)
 
 instance Monoid m => Applicative (Under m) where
     pure _              = Under mempty
     Under x <*> Under y = Under (mappend x y)
 
+-- select = selectT
 instance Monoid m => Selective (Under m) where
     select (Under m) _ = Under m
 
@@ -392,7 +429,7 @@ instance Selective ZipList where select = selectA
 
 -- | Selective instance for the standard applicative functor Validation. This is
 -- a good example of a non-trivial selective functor which is not a monad.
-data Validation e a = Failure e | Success a deriving (Functor, Show)
+data Validation e a = Failure e | Success a deriving (Eq, Functor, Ord, Show)
 
 instance Semigroup e => Applicative (Validation e) where
     pure = Success
@@ -409,6 +446,27 @@ instance Semigroup e => Selective (Validation e) where
 instance (Selective f, Selective g) => Selective (Product f g) where
     select (Pair fx gx) (Pair fy gy) = Pair (select fx fy) (select gx gy)
 
+instance Selective f => Selective (IdentityT f) where
+    select (IdentityT x) (IdentityT y) = IdentityT (select x y)
+
+instance Selective f => Selective (ReaderT env f) where
+    select (ReaderT x) (ReaderT y) = ReaderT $ \env -> select (x env) (y env)
+
+distributeEither :: (Either a b, w) -> Either (a, w) (b, w)
+distributeEither (Left  a, w) = Left  (a, w)
+distributeEither (Right b, w) = Right (b, w)
+
+distributeFunction :: Monoid w => (a -> b, w) -> (a, w) -> (b, w)
+distributeFunction (f, wf) (x, wx) = (f x, mappend wx wf)
+
+instance (Monoid w, Selective f) => Selective (WriterT w f) where
+    select (WriterT x) (WriterT f) =
+        WriterT $ select (distributeEither <$> x) (distributeFunction <$> f)
+
+instance (Monoid w, Selective f) => Selective (S.WriterT w f) where
+    select (S.WriterT x) (S.WriterT f) =
+        S.WriterT $ select (distributeEither <$> x) (distributeFunction <$> f)
+
 -- TODO: Is this a useful instance? Note that composition of 'Alternative'
 -- requires @f@ to be 'Alternative', and @g@ to be 'Applicative', which is
 -- opposite to what we have here:
@@ -453,15 +511,11 @@ instance             Selective STM        where select = selectM
 
 instance                        Selective (ContT      r m) where select = selectM
 instance            Monad m  => Selective (ExceptT    e m) where select = selectM
-instance            Monad m  => Selective (IdentityT    m) where select = selectM
 instance            Monad m  => Selective (MaybeT       m) where select = selectM
-instance            Monad m  => Selective (ReaderT    r m) where select = selectM
 instance (Monoid w, Monad m) => Selective (RWST   r w s m) where select = selectM
 instance (Monoid w, Monad m) => Selective (S.RWST r w s m) where select = selectM
 instance            Monad m  => Selective (StateT     s m) where select = selectM
 instance            Monad m  => Selective (S.StateT   s m) where select = selectM
-instance (Monoid w, Monad m) => Selective (WriterT    w m) where select = selectM
-instance (Monoid w, Monad m) => Selective (S.WriterT  w m) where select = selectM
 
 ------------------------------------ Arrows ------------------------------------
 -- See the following standard definitions in "Control.Arrow".
@@ -473,7 +527,7 @@ instance ArrowChoice a => Selective (ArrowMonad a) where
     select (ArrowMonad x) y = ArrowMonad $ x >>> (toArrow y ||| returnA)
 
 toArrow :: Arrow a => ArrowMonad a (i -> o) -> a i o
-toArrow (ArrowMonad f) = arr (\x -> ((), x)) >>> first f >>> arr (uncurry ($))
+toArrow (ArrowMonad f) = arr ((),) >>> first f >>> arr (uncurry ($))
 
 ---------------------------------- Alternative ---------------------------------
 -- | Composition of a functor @f@ with the 'Either' monad.
